unit TMSRemoteUnit2;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, ExtCtrls, OverbyteIcsWndControl, OverbyteIcsWSocket,
  ImgList, Profiles, CyTools;

type
  TForm2 = class(TForm)
    var
      ListView1        : TListView;
      Timer1           : TTimer;
      WSocket_XMT      : TWSocket;
      WSocket_RCV      : TWSocket;
      ImageList1       : TImageList;

    procedure InitListView();
    procedure Timer1Timer(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure SetMultiInstance(MultiInstance: boolean);

    procedure WSocket_XMTDataAvailable(Sender: TObject; ErrCode: Word);
    procedure WSocket_XMTError(Sender: TObject);
    procedure WSocket_XMTBgException(Sender: TObject; E: Exception; var CanClose: Boolean);
    procedure WSocket_XMTChangeState(Sender: TObject; OldState, NewState: TSocketState);
    procedure WSocket_XMTLineLimitExceeded(Sender: TObject; RcvdLength: Integer; var ClearData: Boolean);
    procedure WSocket_XMTSendData(Sender: TObject; BytesSent: Integer);
    procedure WSocket_XMTSessionAvailable(Sender: TObject; ErrCode: Word);
    procedure WSocket_XMTSessionClosed(Sender: TObject; ErrCode: Word);
    procedure WSocket_XMTSessionConnected(Sender: TObject; ErrCode: Word);
    procedure WSocket_XMTSocksAuthState(Sender: TObject; AuthState: TSocksAuthState);
    procedure WSocket_XMTSocksConnected(Sender: TObject; ErrCode: Word);
    procedure WSocket_XMTSocksError(Sender: TObject; Error: Integer; Msg: string);
    Function  SocketStateString(State: tSocketState): String;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);


  private
    const
      BUFFERSIZE              = 256;

      CMD_NOP                 = $10;
      CMD_Firmware            = $A0;
      CMD_Reboot              = $B0;
      CMD_Factory             = $C0;
      CMD_GetControllerStatus = $D0;

      CMD_SendNECRemote       = $14;
      CMD_SendNECRemote2      = $15;
      CMD_GetToppyStatus      = $24;
      CMD_GetToppyStatusMAC   = $2a;

      STATUS_OK               = $00;
      STATUS_UnknownCommand   = $01;
      STATUS_InvalidParameter = $02;
      STATUS_ReqSettings      = $0f;

    type
      tConnectState = (CONNECTSTATE_Init,
                       CONNECTSTATE_SendStatusRequest,
                       CONNECTSTATE_WaitForStatusResponse,
                       CONNECTSTATE_StatusResponseReceived,
                       CONNECTSTATE_TryConnectTMSRemote,
                       CONNECTSTATE_WaitTMSRemote,
                       CONNECTSTATE_Exit,
                       CONNECTSTATE_Invalid);

    var
      ConnectState        : tConnectState;
      LastConnectState    : tConnectState;
      RetryCount          : Integer;
      PowerRetryCount     : Integer;
      Timeout             : dword;
      LastPowerOffTimeout : dword;
      ConnectionTimeout   : dword;
      SEQ                 : Byte;
      ToppyStatusSEQ      : Integer;
      PowerStatus         : Boolean;
      LastSocketState     : TSocketState;
      CP                  : ^tConnectionProfile;
      LastTimeoutSeconds  : Integer;
      TMSRMultiInstance   : Boolean;

    procedure SendStatusRequest(IPAddr: pByteArray);
    procedure SendStatusRequestMAC(IPAddr: pByteArray; MACAddr: pByteArray);
    procedure SendPowerOn();

  public
    procedure SendIRCode(IRCode: word; UsePowerOnMode: Boolean);

  end;

var
  Form2: TForm2;

implementation

uses TMSRemoteUnit1;

{$R *.dfm}

procedure TForm2.FormShow(Sender: TObject);
begin
  InitListView();

  CP := @ConnectionProfile[CurrentConnectionProfile];
  ConnectState := CONNECTSTATE_Init;
  LastConnectState := CONNECTSTATE_Invalid;
  LastSocketState := wsInvalidState;
  LastPowerOffTimeout := 0;
  Timer1.Enabled := true;
end;

procedure TForm2.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if ConnectState <> CONNECTSTATE_Exit then
  begin
    Timer1.Enabled := false;
    WSocket_XMT.Close();
    WSocket_RCV.Close();
    Self.ModalResult := mrAbort;
  end;
end;

procedure TForm2.SetMultiInstance(MultiInstance: boolean);
begin
  TMSRMultiInstance := MultiInstance;
end;

procedure TForm2.InitListView();

var Col                      : TListColumn;
    Item                     : tListItem;

begin
  with ListView1 do
  begin
    ViewStyle := vsReport;
    Checkboxes := false;
    GridLines := true;
    ReadOnly := true;
    RowSelect := false;
    SmallImages := ImageList1;

    with Columns do
    begin
      Clear();
      Col           := Add();
      Col.Caption   := 'Action';
      Col.Alignment := taLeftJustify;
      Col.Width     := 240;
    end;

    Items.Clear();
    Font.Size := 10;
    Item := Items.Add();
    Item.Caption := '  Waiting for TFIR';
    Item.ImageIndex := -1;

    Item := Items.Add();
    Item.Caption := '  Waiting for Toppy';
    Item.ImageIndex := -1;

    Item := Items.Add();
    Item.Caption := '  Waiting for TMSRemote';
    Item.ImageIndex := -1;
  end;
end;

procedure TForm2.Timer1Timer(Sender: TObject);

var
  dTimeout             : integer;
  IPAddress            : array[0..3] of byte;
  MACAddress           : array[0..5] of byte;
  i                    : Integer;
  d                    : dword;

begin
  if ConnectState <> LastConnectState then
  begin
    case ConnectState of
      CONNECTSTATE_Init:                   OutputDebugString('CONNECTSTATE_Init');
      CONNECTSTATE_SendStatusRequest:      OutputDebugString('CONNECTSTATE_SendStatusRequest');
      CONNECTSTATE_WaitForStatusResponse:  OutputDebugString('CONNECTSTATE_WaitForStatusResponse');
      CONNECTSTATE_StatusResponseReceived: OutputDebugString('CONNECTSTATE_StatusResponseReceived');
      CONNECTSTATE_TryConnectTMSRemote:    OutputDebugString('CONNECTSTATE_TryConnectTMSRemote');
      CONNECTSTATE_WaitTMSRemote:          OutputDebugString('CONNECTSTATE_WaitTMSRemote');
      CONNECTSTATE_Exit:                   OutputDebugString('CONNECTSTATE_Exit');
      CONNECTSTATE_Invalid:                OutputDebugString('CONNECTSTATE_Invalid');
    end;
    LastConnectState := ConnectState;
  end;

  case ConnectState of
    CONNECTSTATE_Init:
    begin
      if CP.TFIRInUse and not(TMSRMultiInstance) then
      begin
        WSocket_XMT.Addr := CP.TFIRIP;
        WSocket_XMT.Port := format('%d', [CP.TFIRPort]);
        WSocket_XMT.LocalPort := WSocket_XMT.Port;
        WSocket_XMT.Proto := 'udp';
        try
          WSocket_XMT.Connect();
        except

        end;

        WSocket_RCV.Close();
        WSocket_RCV.Addr := CP.TFIRIP;
        WSocket_RCV.Port := WSocket_XMT.Port;
        WSocket_RCV.LocalPort := WSocket_XMT.Port;
        WSocket_RCV.Proto := 'udp';
        try
          WSocket_RCV.Listen();
        except

        end;

        RetryCount := 0;
        PowerRetryCount := 0;

        ConnectState := CONNECTSTATE_SendStatusRequest;
      end else
      begin
        ListView1.Items[0].ImageIndex := 1;
        ListView1.Items[1].ImageIndex := 1;
        Listview1.Refresh();
        ConnectionTimeout := GetTickCount() + CP.TMSRemoteTimeout *  1000;
        ConnectState := CONNECTSTATE_TryConnectTMSRemote
      end;
    end;

    CONNECTSTATE_SendStatusRequest:
    begin
      inc(RetryCount);
      if RetryCount > CP.TFIRRetries then
      begin
        OutputDebugString('No response from TFIR');
        Timer1.Enabled := false;
        WSocket_XMT.Close();
        WSocket_RCV.Close();
        ListView1.Items[0].ImageIndex := 1;
        Listview1.Refresh();
        MessageBox(Self.WindowHandle, 'No response from TFIR', 'TMSRemote', MB_OK);
        Self.ModalResult := mrAbort;
        exit;
      end;

      ConnectState := CONNECTSTATE_WaitForStatusResponse;
      Timeout := GetTickCount() + CP.TFIRTimeout * 1000;

      d := CyTools.StringToIP(CP.TFIRPingIP);
      IPAddress[3] := d shr 24;
      IPAddress[2] := (d shr 16) and $ff;
      IPAddress[1] := (d shr  8) and $ff;
      IPAddress[0] := d and $ff;

      for i := 0 to 5 do
        MACAddress[i] := StrToInt('$' + Copy(CP.TFIRPingMAC, 3*i + 1, 2));
      SendStatusRequestMAC(@IPAddress[0], @MACAddress[0])
    end;

    CONNECTSTATE_WaitForStatusResponse:
    begin
      if GetTickCount() > Timeout then
      begin
        OutputDebugString('Timeout waiting for TFIR response');
        ConnectState := CONNECTSTATE_SendStatusRequest;
      end;
    end;

    CONNECTSTATE_StatusResponseReceived:
    begin
      ListView1.Items[0].ImageIndex := 0;
      if PowerStatus then
      begin
        OutputDebugString('Toppy is ON');
        ListView1.Items[1].ImageIndex := 0;
        ConnectionTimeout := GetTickCount() + CP.TMSRemoteTimeout *  1000;
        ConnectState := CONNECTSTATE_TryConnectTMSRemote;
      end else
      begin
        OutputDebugString('Toppy is OFF');

        if LastPowerOffTimeout > 0 then
        begin
          dTimeout := Trunc((LastPowerOffTimeout - GetTickCount()) / 1000);
          if dTimeout < 0 then dTimeout := 0;
          if dTimeout <> LastTimeoutSeconds then
          begin
            ListView1.Items[1].Caption := Format('  Waiting for Toppy (%d")', [dTimeout]);
            LastTimeoutSeconds := dTimeout;
          end;
        end;

        if GetTickCount() > LastPowerOffTimeout then
        begin
          inc(PowerRetryCount);
          if PowerRetryCount > CP.BootRetries then
          begin
            OutputDebugString('Failed to start Toppy');
            Timer1.Enabled := false;
            WSocket_XMT.Close();
            WSocket_RCV.Close();
            ListView1.Items[1].ImageIndex := 1;
            ListView1.Refresh();
            MessageBox(Self.WindowHandle, 'Failed to start Toppy', 'TMSRemote', MB_OK);
            Self.ModalResult := mrAbort;
            exit;
          end;
          OutputDebugString('Sending power on command');
          SendPowerOn();
          LastPowerOffTimeout := GetTickCount() + CP.BootTimeout * 1000;
        end;
        RetryCount := 0;
        ConnectState := CONNECTSTATE_SendStatusRequest;
      end;
    end;

    CONNECTSTATE_TryConnectTMSRemote:
    begin
      WSocket_XMT.Close();
      WSocket_RCV.Close();

      dTimeout := Trunc((ConnectionTimeout - GetTickCount()) / 1000);
      if dTimeout < 0 then dTimeout := 0;
      if dTimeout <> LastTimeoutSeconds then
      begin
        ListView1.Items[2].Caption := Format('  Waiting for TMSRemote (%d")', [dTimeout]);
        LastTimeoutSeconds := dTimeout;
      end;

      if GetTickCount() > ConnectionTimeout then
      begin
        if CP.TFIRInUse then
          OutputDebugString('Toppy is on but TMSRemote did not respond')
        else
          OutputDebugString('TMSRemote did not respond');
        Timer1.Enabled := false;
        WSocket_XMT.Close();
        WSocket_RCV.Close();
        ListView1.Items[2].ImageIndex := 1;
        ListView1.Refresh();
        if Cp.TFIRInUse then
          MessageBox(Self.WindowHandle, 'Toppy is on but TMSRemote does not respond', 'TMSRemote', MB_OK)
        else
          MessageBox(Self.WindowHandle, 'Toppy is not reachable (powered off or TMSRemote is not responding)', 'TMSRemote', MB_OK);
        Self.ModalResult := mrAbort;
        exit;
      end;

      WSocket_XMT.Addr := CP.ToppyIP;
      WSocket_XMT.Port := format('%d', [CP.ToppyPort]);
      WSocket_XMT.Proto := 'tcp';

      ConnectState := CONNECTSTATE_WaitTMSRemote;
      try
        WSocket_XMT.Connect;

      except

      end;
    end;

    CONNECTSTATE_WaitTMSRemote:
    begin
      if WSocket_XMT.State <>  LastSocketState then
      begin
        case WSocket_XMT.State of
          wsInvalidState:   OutputDebugString('wsInvalidState');
          wsOpened:         OutputDebugString('wsOpened');
          wsBound:          OutputDebugString('wsBound');
          wsConnecting:     OutputDebugString('wsConnecting');
          wsSocksConnected: OutputDebugString('wsSocksConnected');
          wsConnected:      OutputDebugString('wsConnected');
          wsAccepting:      OutputDebugString('wsAccepting');
          wsListening:      OutputDebugString('wsListening');
          wsClosed:         OutputDebugString('wsClosed');
        end;
        LastSocketState := WSocket_XMT.State;
      end;

      if WSocket_XMT.State = wsConnected then
      begin
        OutputDebugString('Sucessfully connected, returning to TMSRemote');
        ListView1.Items[2].ImageIndex := 0;
        ListView1.Refresh();
        WSocket_XMT.Close();
        WSocket_RCV.Close();
        ConnectState := CONNECTSTATE_Exit;
        exit;
      end;

      if WSocket_XMT.State = wsClosed then
      begin
        ConnectState := CONNECTSTATE_TryConnectTMSRemote;
      end;

    end;

    CONNECTSTATE_Exit:
    begin
      Timer1.Enabled := false;
      Self.ModalResult := mrOk;
    end;
  end;
end;

procedure TForm2.SendStatusRequest(IPAddr: pByteArray);

var
  Data                 : Array[0..7] of byte;

begin
  inc(SEQ);
  ToppyStatusSEQ := SEQ;

  Data[0] := $02;
  Data[1] := SEQ;
  Data[2] := CMD_GetToppyStatus;

  Data[3] := IPAddr[0];
  Data[4] := IPAddr[1];
  Data[5] := IPAddr[2];
  Data[6] := IPAddr[3];

  Data[7] := $03;
  WSocket_XMT.Send(@Data[0], 8);
end;

procedure TForm2.SendStatusRequestMAC(IPAddr: pByteArray; MACAddr: pByteArray);

var
  Data                 : Array[0..13] of byte;

begin
  inc(SEQ);
  ToppyStatusSEQ := SEQ;

  Data[0] := $02;
  Data[1] := SEQ;
  Data[2] := CMD_GetToppyStatusMAC;

  Data[3] := IPAddr[0];
  Data[4] := IPAddr[1];
  Data[5] := IPAddr[2];
  Data[6] := IPAddr[3];

  Data[7] := MACAddr[0];
  Data[8] := MACAddr[1];
  Data[9] := MACAddr[2];
  Data[10] := MACAddr[3];
  Data[11] := MACAddr[4];
  Data[12] := MACAddr[5];

  Data[13] := $03;
  WSocket_XMT.Send(@Data[0], 14);
end;

procedure TForm2.SendPowerOn();
begin
  SendIRCode($50af, true);
end;

procedure TForm2.SendIRCode(IRCode: word; UsePowerOnMode: Boolean);

var
  Data                 : Array[0..8] of byte;
  IRMode               : byte;

begin
  inc(SEQ);

  Data[0] := $02;
  Data[1] := SEQ;
  Data[2] := CMD_SendNECRemote2;
  Data[3] := $00;

  if UsePowerOnMode then IRMode := CP.TFIRPowerOnMode
                    else IRMode := CP.TFIRRemoteMode;
  if IRMode = 0 then IRMode := CP.TFIRRemoteMode;

  case IRMode of
    1:
    begin
      Data[4] := $40;
      Data[5] := $2c;
    end;

    2:
    begin
      Data[4] := $92;
      Data[5] := $00;
    end;

    3:
    begin
      Data[4] := $92;
      Data[5] := $99;
    end;

    4:
    begin
      Data[4] := $04;
      Data[5] := $fb;
    end;

    5:
    begin
      Data[4] := $0a;
      Data[5] := $80;
    end;
  end;

  Data[6] := IRCode shr 8;
  Data[7] := IRCode and $ff;
  Data[8] := $03;

  WSocket_XMT.Send(@Data[0], 9);
end;

procedure TForm2.WSocket_XMTBgException(Sender: TObject; E: Exception; var CanClose: Boolean);
begin
  OutputDebugString(pchar(Format('WSocket_XMTBgException: %s', [E.Message])));
end;

Function TForm2.SocketStateString(State: tSocketState): String;
begin
  case State of
    wsInvalidState:    Result := 'wsInvalidState';
    wsOpened:          Result := 'wsOpened';
    wsBound:           Result := 'wsBound';
    wsConnecting:      Result := 'wsConnecting';
    wsSocksConnected:  Result := 'wsSocksConnected';
    wsConnected:       Result := 'wsConnected';
    wsAccepting:       Result := 'wsAccepting';
    wsListening:       Result := 'wsListening';
    wsClosed:          Result := 'wsClosed';
  end;
end;

procedure TForm2.WSocket_XMTChangeState(Sender: TObject; OldState, NewState: TSocketState);
begin
  OutputDebugString(pChar(Format('WSocket_XMTChangeState: %s -> %s', [SocketStateString(OldState), SocketStateString(NewState)])));
end;

procedure TForm2.WSocket_XMTDataAvailable(Sender: TObject; ErrCode: Word);

var
  Count                : Integer;
  Socket               : TWSocket;
  Buffer               : array [0..BUFFERSIZE-1] of byte;

begin
  OutputDebugString(pChar(Format('WSocket_XMTDataAvailable: ErrCode=%d', [ErrCode])));

  if ErrCode <> 0 then
  begin
    OutputDebugString(pchar(format('Network: XMT DataAvailable. ErrCode = 0x%x', [ErrCode])));
    Exit;
  end;

  Socket := Sender as TWSocket;
  Count := Socket.Receive(@Buffer[0], BUFFERSIZE);
  if Count > 0 then
  begin
    if (Buffer[0] = $02) and (Buffer[1] = ToppyStatusSEQ) then
    begin
      PowerStatus := (Buffer[4] and $01) <> 0;
      ToppyStatusSEQ := -1;
      ConnectState := CONNECTSTATE_StatusResponseReceived;
    end;
  end;
end;

procedure TForm2.WSocket_XMTError(Sender: TObject);
begin
  with Sender as TWSocket do
    OutputDebugString(pChar(Format('WSocket_XMTError: %d = %s', [LastError, WsocketErrorDesc(LastError)])));

  Timer1.Enabled := false;
  ConnectState := CONNECTSTATE_Exit;
  MessageBox(Self.WindowHandle, 'Network error', 'AutoConnector', MB_OK);
  WSocket_XMT.Close();
  Self.ModalResult := mrCancel;
end;

procedure TForm2.WSocket_XMTLineLimitExceeded(Sender: TObject; RcvdLength: Integer; var ClearData: Boolean);
begin
  OutputDebugString('WSocket_XMTLineLimitExceeded');
end;

procedure TForm2.WSocket_XMTSendData(Sender: TObject; BytesSent: Integer);
begin
  OutputDebugString('WSocket_XMTSendData');
end;

procedure TForm2.WSocket_XMTSessionAvailable(Sender: TObject; ErrCode: Word);
begin
  OutputDebugString('WSocket_XMTSessionAvailable');
end;

procedure TForm2.WSocket_XMTSessionClosed(Sender: TObject; ErrCode: Word);
begin
  OutputDebugString('WSocket_XMTSessionClosed');
end;

procedure TForm2.WSocket_XMTSessionConnected(Sender: TObject; ErrCode: Word);
begin
  OutputDebugString('WSocket_XMTSessionConnected');
end;

procedure TForm2.WSocket_XMTSocksAuthState(Sender: TObject; AuthState: TSocksAuthState);
begin
  OutputDebugString('WSocket_XMTSocksAuthState');
end;

procedure TForm2.WSocket_XMTSocksConnected(Sender: TObject; ErrCode: Word);
begin
  OutputDebugString('WSocket_XMTSocksConnected');
end;

procedure TForm2.WSocket_XMTSocksError(Sender: TObject; Error: Integer; Msg: string);
begin
  OutputDebugString('WSocket_XMTSocksError');
end;

end.

