unit TMSRemoteVFDUnit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, OverbyteIcsWndControl, OverbyteIcsWSocket, StdCtrls, Registry,
  jpeg, CyTools, Menus, ExtDlgs, INIFiles, XiProgressBar, IdBaseComponent,
  IdComponent, IdRawBase, IdRawClient, IdIcmpClient, Profiles, Vcl.Imaging.pngimage;

const
  ROOTKEY              = 'Software\FireBird\TMSRemote';
  DEFAULTBUFFERSIZE    = $100000;

type

  TForm1 = class(TForm)
    WSocket1                 : TWSocket;
    Button_Connect           : TButton;
    Label_IP                 : TLabel;
    MainMenu1                : TMainMenu;
    Menu_File                : TMenuItem;
    Menu_File_Sep1           : TMenuItem;
    Menu_File_Exit           : TMenuItem;
    ComboBox_Profile         : TComboBox;
    Menu_File_SaveScreen     : TMenuItem;
    Timer1                   : TTimer;
    IdIcmpClient1            : TIdIcmpClient;
    Image_VFD                : TImage;
    SavePictureDialog1       : TSavePictureDialog;
    Image1                   : TImage;
    Timer2                   : TTimer;
    Menu_File_NewInstance    : TMenuItem;
    TrayIcon1                : TTrayIcon;
    File_Menu_MinimizeToTray : TMenuItem;

    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure WSocket1DataAvailable(Sender: TObject; ErrCode: Word);
    procedure Button_ConnectClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure WSocket1SessionConnected(Sender: TObject; ErrCode: Word);
    procedure WSocket1SessionClosed(Sender: TObject; ErrCode: Word);
    procedure Menu_File_SaveScreenClick(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure Menu_File_ExitClick(Sender: TObject);
    procedure Timer2Timer(Sender: TObject);
    procedure ComboBox_ProfileChange(Sender: TObject);
    procedure Menu_File_NewInstanceClick(Sender: TObject);
    procedure TrayIcon1Click(Sender: TObject);
    procedure File_Menu_MinimizeToTrayClick(Sender: TObject);
    procedure FormResize(Sender: TObject);

  private
    var
      CommConnected          : Boolean;
      AliveTimer             : dword;
      DoNotReenterDataAvail  : boolean;
      Grid                   : Array[0..47] of byte;
      t                      : boolean;
      hSemaphore             : dword;
      OfflineCaption         : String;
      ScreenshotFormat       : String;

    //GUI
    procedure ClearVFD();
    procedure UpdateVFD(NewGrid: pByteArray);
    procedure LoadVFDIcon(Name: String; HasMask: Boolean; x, y: dword; OldActiveByte, CurrentActiveByte, Bit: byte);

    //Settings
    procedure LoadSettings();
    procedure SaveSettings();
    procedure UpdateConnectionProfileListbox();

    //Network
    procedure CommOpen();
    procedure CommClose();
    procedure SendAlive();

  public

  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}
{$R 'VFD Icons\TMSRemote_Icons.res'}

{$REGION 'Settings'}
procedure TForm1.LoadSettings();

var
  Reg                  : tRegistry;
  i                    : Integer;
  HideOsd              : boolean;
  INI                  : tINIFile;

begin
  //Make sure that there is no combobox event triggered or there may be a save settings before all settings have been read
  if DoesINIProfileExist() then
  begin
    INI := tINIFile.Create(GetININame());

    HideOsd := INI.ReadBool('Window', 'HideOsd', FALSE);
    ScreenshotFormat := INI.ReadString('Window', 'ScreenshotFormat', '');

    //Read all connection profiles
    i := 0;
    while True do
    begin
      if not(ReadConnectionProfile(i, ConnectionProfile[i])) then break;
      inc(i);
    end;
    NrConnectionProfiles := i;

    if NrConnectionProfiles = 0 then
    begin
      MessageBox(Self.WindowHandle, 'At least one connection profile is necessary.'#13#10'Please restart TMSRemote and define a profile.', 'TMSRemote', MB_OK);
      halt(1);
    end;

    if INI.ValueExists('Window', 'CurrentConnectionProfile') then
      CurrentConnectionProfile := INI.ReadInteger('Window', 'CurrentConnectionProfile', 1) - 1
    else
      CurrentConnectionProfile := 0;

    File_Menu_MinimizeToTray.Checked := INI.ReadBool('Window', 'MinimizeToTray', false);

    INI.Free();
  end else
  begin
    Reg := tRegistry.Create();
    Reg.RootKey := HKEY_CURRENT_USER;

    Reg.OpenKey(ROOTKEY, true);

    HideOsd := false;
    if Reg.ValueExists('HideOsd') then
      HideOsd := Reg.ReadBool('HideOsd');

    if Reg.ValueExists('ScreenshotFormat') then
      ScreenshotFormat := Reg.ReadString('ScreenshotFormat')
    else
      ScreenshotFormat := '';

    //Delete all old entries
    i := 0;
    while True do
    begin
      inc(i);
      if not(Reg.ValueExists(Format('IP%d', [i]))) then break;
      Reg.DeleteValue(Format('IP%d', [i]));
    end;
    if Reg.ValueExists('tcpPort') then Reg.DeleteValue('tcpPort');

    //Read all connection profiles
    i := 0;
    while True do
    begin
      if not(ReadConnectionProfile(i, ConnectionProfile[i])) then break;
      inc(i);
    end;
    NrConnectionProfiles := i;

    if NrConnectionProfiles = 0 then
    begin
      MessageBox(Self.WindowHandle, 'At least one connection profile is necessary.'#13#10'Please restart TMSRemote and define a profile.', 'TMSRemote', MB_OK);
      halt(1);
    end;

    if Reg.ValueExists('CurrentConnectionProfile') then
      CurrentConnectionProfile := Reg.ReadInteger('CurrentConnectionProfile') - 1
    else
      CurrentConnectionProfile := 0;

    if Reg.ValueExists('MinimizeToTray') then
      File_Menu_MinimizeToTray.Checked := Reg.ReadBool('MinimizeToTray');

    Reg.CloseKey();
    Reg.Free();
  end;

  UpdateConnectionProfileListbox();
  ComboBox_Profile.ItemIndex := CurrentConnectionProfile;
end;

procedure TForm1.SaveSettings();

var
  Reg                  : tRegistry;
  INI                  : tINIFile;
  i                    : Integer;

begin
  //Write all connection profiles
  for i := 0 to NrConnectionProfiles - 1 do
    WriteConnectionProfile(i, ConnectionProfile[i]);

  if DoesINIProfileExist() then
  begin
    if (GetFileAttributes(pChar(GetININame())) and FILE_ATTRIBUTE_READONLY) = 0 then
    begin
      INI := tINIFile.Create(GetININame());
      INI.WriteInteger('Window', 'CurrentConnectionProfile', CurrentConnectionProfile + 1);
      INI.WriteBool('Window', 'MinimizeToTray', File_Menu_MinimizeToTray.Checked);
      INI.WriteString('Window', 'ScreenshotFormat', ScreenshotFormat);
      INI.Free();
    end;
  end else
  begin
    Reg := tRegistry.Create();
    Reg.RootKey := HKEY_CURRENT_USER;

    Reg.OpenKey(ROOTKEY, true);
    Reg.WriteString('ScreenshotFormat', ScreenshotFormat);
    Reg.WriteInteger('CurrentConnectionProfile', CurrentConnectionProfile + 1);
    Reg.WriteBool('MinimizeToTray', File_Menu_MinimizeToTray.Checked);
    Reg.CloseKey();
    Reg.Free();
  end;
end;

procedure TForm1.UpdateConnectionProfileListbox();

var
  i, j                 : Integer;

begin
  j := ComboBox_Profile.ItemIndex;
  ComboBox_Profile.Items.Clear();
  for i := 0 to NrConnectionProfiles - 1 do
    ComboBox_Profile.Items.Add(ConnectionProfile[i].ProfileName);
  ComboBox_Profile.ItemIndex := j;
end;
{$ENDREGION}

{$REGION 'GUI'}
procedure TForm1.FormCreate(Sender: TObject);

var
  hWnd                  : dword;
  hSemaphore            : dword;
  NewInstance           : Boolean;
  i                     : integer;

begin
  OfflineCaption := 'TMS Remote VFD - V' + CyTools.GetProductVersion(paramstr(0));
  NewInstance := false;

  for i := 1 to ParamCount do
  begin
    if LowerCase(Copy(ParamStr(i), 1, 12)) = '/newinstance' then NewInstance := true;
  end;

  if not(NewInstance) then
  begin
    Form1.Caption := 'temp delete';
    hSemaphore := CreateSemaphore(nil, 1, 1, 'TMSRemoteVFDSemaphore');
    if hSemaphore = 0 then halt(1);
    if WaitForSingleObject(hSemaphore, 100) <> WAIT_OBJECT_0 then
    begin
      CloseHandle(hSemaphore);
      hWnd := FindWindow(nil, pChar(OfflineCaption));
      if hWnd <> 0 then ForceForegroundWindow(hWnd);
      halt(2);
    end;
    Form1.Caption := OfflineCaption;
  end;

  TrayIcon1.Icon := Form1.Icon;
  TrayIcon1.Hint := 'TMSVFD';
  TrayIcon1.Visible := true;

  Form1.Top := 0;
  DoNotReenterDataAvail := false;
  LoadSettings();
  Image_VFD.Picture.Bitmap.LoadFromResourceName(HInstance, 'VFD_Off');
  FillMemory(@Grid[0], 48, $ff);
  ClearVFD();
end;

procedure TForm1.FormResize(Sender: TObject);
begin
  if File_Menu_MinimizeToTray.Checked and (Form1.WindowState = wsMinimized) then
    Form1.Hide();
end;

procedure TForm1.File_Menu_MinimizeToTrayClick(Sender: TObject);
begin
  File_Menu_MinimizeToTray.Checked := not(File_Menu_MinimizeToTray.Checked);
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  WSocket1.Close();
  ReleaseSemaphore(hSemaphore, 1, nil);
  TrayIcon1.Visible := false;
  SaveSettings();
end;

procedure TForm1.Button_ConnectClick(Sender: TObject);
begin
  Button_Connect.Enabled := false;
  if CommConnected then
  begin
    CommClose();
    Form1.Caption := OfflineCaption;
    TrayIcon1.Hint := 'TMSVFD';
  end else
  begin
    CommOpen();
    Form1.Caption := ComboBox_Profile.Items[ComboBox_Profile.ItemIndex] + ' - TMS Remote VFD';
    TrayIcon1.Hint := ComboBox_Profile.Items[ComboBox_Profile.ItemIndex];
    SaveSettings();
  end;
end;

procedure TForm1.ComboBox_ProfileChange(Sender: TObject);
begin
  CurrentConnectionProfile := ComboBox_Profile.ItemIndex;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  SendAlive();
end;

procedure TForm1.Timer2Timer(Sender: TObject);

var
  TempGrid             : array[0..47] of byte;

begin
  if t then
  begin
    FillMemory(@TempGrid[0], 48, $ff);
    UpdateVFD(@TempGrid[0]);
  end else
    ClearVFD();

  t := not(t);
end;

procedure TForm1.TrayIcon1Click(Sender: TObject);
begin
  Form1.WindowState := wsNormal;
  Form1.Show;
  SetWindowPos(Form1.Handle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE or SWP_NOSIZE);
  Application.ProcessMessages;
  SetWindowPos(Form1.Handle, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE or SWP_NOSIZE);
end;

procedure TForm1.Menu_File_ExitClick(Sender: TObject);
begin
  Self.Close();
end;

procedure TForm1.Menu_File_NewInstanceClick(Sender: TObject);

var
  TMSRemoteVFDEXE       : String;
  RC                    : dword;
  ProcessInfo           : tProcessInformation;

begin
  TMSRemoteVFDEXE := Paramstr(0);
  CyTools.ExecuteFileAndWait(TMSRemoteVFDEXE, '/NewInstance', ExtractFilePath(TMSRemoteVFDEXE), 1, false, RC, ProcessInfo);
end;

procedure TForm1.Menu_File_SaveScreenClick(Sender: TObject);

var
  jpg                  : tJPEGImage;
  png                  : TPNGObject;
  FileName, Ext        : String;
  i                    : integer;
  SeparatorChar        : Array[0..0] of Char;
  FilterList           : tArray<string>;

begin
  SavePictureDialog1.Filter := 'JPEG (*.jpg)|*.jpg|Bitmap (*.bmp)|*.bmp|PNG (*.png)|*.png|All files|*.*';

  if ScreenshotFormat <> '' then
  begin
    SavePictureDialog1.DefaultExt := ScreenshotFormat;

    //Adapt FilterIndex to the last file extension
    SeparatorChar[0] := '|';
    FilterList := SavePictureDialog1.Filter.Split(SeparatorChar);
    for i := 0 to High(FilterList) do
      if pos(ScreenshotFormat, FilterList[i]) > 0 then
      begin
        SavePictureDialog1.FilterIndex := (i shr 1) + 1;
        break;
      end;
  end else
  begin
    SavePictureDialog1.DefaultExt := '.jpg';
    SavePictureDialog1.FilterIndex := 1;
  end;

  //Display the dialog
  if SavePictureDialog1.Execute then
  begin
    FileName := SavePictureDialog1.FileName;

    //If there's still no file name extension, make sure to add one according to the dialog's filter settings
    if ExtractFileExt(FileName) = '' then
      case SavePictureDialog1.FilterIndex of
        1: FileName := ChangeFileExt(FileName, '.jpg');
        2: FileName := ChangeFileExt(FileName, '.bmp');
        3: FileName := ChangeFileExt(FileName, '.png');
      end;

    Ext := LowerCase(ExtractFileExt(FileName));
    ScreenshotFormat := Ext;
    SaveSettings();

    if Ext = '.bmp' then
      Image_VFD.Picture.SaveToFile(FileName);

    if Ext = '.jpg' then
    begin
      jpg := tJPEGImage.Create();
      jpg.Performance := jpBestQuality;
      jpg.Assign(Image_VFD.Picture.Bitmap);
      jpg.SaveToFile(FileName);
      jpg.Free();
    end;

    if Ext = '.png' then
    begin
      png := TPNGObject.Create;
      png.Assign(Image_VFD.Picture.Bitmap);
      png.SaveToFile(FileName);
      png.Free;
    end;
  end;
end;

procedure TForm1.ClearVFD();

var
  TempGrid             : Array[0..47] of byte;

begin
  ZeroMemory(@TempGrid[0], 48);
  UpdateVFD(@TempGrid[0]);
end;

procedure TForm1.LoadVFDIcon(Name: String; HasMask: Boolean; x, y: dword; OldActiveByte, CurrentActiveByte, Bit: byte);

var
  SourceRect           : tRect;
  DestRect             : tRect;

begin
  if (OldActiveByte and (1 shl Bit)) = (CurrentActiveByte and (1 shl Bit)) then exit;

  SourceRect.TopLeft     := Point(0, 0);
  DestRect.TopLeft       := Point(x, y);

  if HasMask then
  begin
    Image1.Picture.Bitmap.LoadFromResourceName(HInstance, Name + '_Mask');

    SourceRect.BottomRight := Point(Image1.Width, Image1.Height);
    DestRect.BottomRight   := Point(x + Image1.Width, y + Image1.Height);

    Image_VFD.Canvas.CopyMode := cmSrcAnd;
    Image_VFD.Canvas.CopyRect(DestRect, Image1.Canvas, SourceRect);
  end;

  if (CurrentActiveByte and (1 shl Bit)) = 0 then
    Image1.Picture.Bitmap.LoadFromResourceName(HInstance, Name + '_Off')
  else
    Image1.Picture.Bitmap.LoadFromResourceName(HInstance, Name + '_On');

  SourceRect.BottomRight := Point(Image1.Width, Image1.Height);
  DestRect.BottomRight   := Point(x + Image1.Width, y + Image1.Height);

  if HasMask then
    Image_VFD.Canvas.CopyMode := cmSrcPaint
  else
    Image_VFD.Canvas.CopyMode := cmSrcCopy;

  Image_VFD.Canvas.CopyRect(DestRect, Image1.Canvas, SourceRect);
end;

procedure TForm1.UpdateVFD(NewGrid: pByteArray);

//var
//  SourceRect, DestRect: tRect;

begin
  LoadVFDIcon('AC3',        false,  63, 33, Grid[27], NewGrid[27], 0);
  LoadVFDIcon('AM',         false, 286, 89, Grid[39], NewGrid[39], 7);
  LoadVFDIcon('Attention',  false, 693, 11, Grid[11], NewGrid[11], 4);
  LoadVFDIcon('Clock',      false, 292, 69, Grid[39], NewGrid[39], 6);
  LoadVFDIcon('Colon',      true,  201, 79, Grid[37], NewGrid[37], 7);
  LoadVFDIcon('Dolby',      false, 728, 10, Grid[11], NewGrid[11], 3);
  LoadVFDIcon('Dollar',     false, 671, 11, Grid[11], NewGrid[11], 5);
  LoadVFDIcon('FFWD',       false, 533, 11, Grid[10], NewGrid[10], 2);
  LoadVFDIcon('HDD',        false,  16,  9, Grid[29], NewGrid[29], 4);
  LoadVFDIcon('HDDBar',     true,   62, 10, Grid[28], NewGrid[28], 5);
  LoadVFDIcon('HDDDot',     false,  68, 13, Grid[29], NewGrid[29], 5);
  LoadVFDIcon('HDDDot',     false,  79, 13, Grid[29], NewGrid[29], 6);
  LoadVFDIcon('HDDDot',     false,  91, 13, Grid[29], NewGrid[29], 7);
  LoadVFDIcon('HDDDot',     false, 102, 13, Grid[28], NewGrid[28], 0);
  LoadVFDIcon('HDDDot',     false, 113, 13, Grid[28], NewGrid[28], 1);
  LoadVFDIcon('HDDDot',     false, 124, 13, Grid[28], NewGrid[28], 2);
  LoadVFDIcon('HDDDot',     false, 135, 13, Grid[28], NewGrid[28], 3);
  LoadVFDIcon('HDDDot',     false, 146, 13, Grid[28], NewGrid[28], 4);
  LoadVFDIcon('HDDFull',    false, 158,  9, Grid[28], NewGrid[28], 6);
  LoadVFDIcon('MP3',        false,  13, 32, Grid[28], NewGrid[28], 7);
  LoadVFDIcon('Mute',       false, 597, 11, Grid[10], NewGrid[10], 0);
  LoadVFDIcon('Network',    false, 769, 10, Grid[11], NewGrid[11], 2);
  LoadVFDIcon('Next',       false, 500, 10, Grid[10], NewGrid[10], 3);
  LoadVFDIcon('Pause',      false, 575, 10, Grid[10], NewGrid[10], 1);
  LoadVFDIcon('Play',       false, 476, 11, Grid[10], NewGrid[10], 4);
  LoadVFDIcon('PM',         false, 284,103, Grid[38], NewGrid[38], 0);
  LoadVFDIcon('Power',      false, 110, 69, Grid[33], NewGrid[33], 5);
  LoadVFDIcon('Prev',       false, 445, 11, Grid[10], NewGrid[10], 5);
  LoadVFDIcon('Radio',      false, 227, 31, Grid[27], NewGrid[27], 3);
  LoadVFDIcon('REC',        false, 309, 12, Grid[ 9], NewGrid[ 9], 1);
  LoadVFDIcon('RecSlot1',   false, 353, 10, Grid[ 9], NewGrid[ 9], 0);
  LoadVFDIcon('RecSlot2',   false, 377, 10, Grid[10], NewGrid[10], 7);
  LoadVFDIcon('Remote',     false, 111,102, Grid[33], NewGrid[33], 4);
  LoadVFDIcon('RepeatLeft', true,   623, 11, Grid[11], NewGrid[11], 7);
  LoadVFDIcon('RepeatRight',true,638, 11, Grid[11], NewGrid[11], 6);
  LoadVFDIcon('RWD',        false, 404, 11, Grid[10], NewGrid[10], 6);
  LoadVFDIcon('Satellite',  false, 247, 29, Grid[27], NewGrid[27], 4);
  LoadVFDIcon('TimeShift',  false, 117, 33, Grid[27], NewGrid[27], 1);
  LoadVFDIcon('TV',         false, 204, 28, Grid[27], NewGrid[27], 2);

  //14 Segment, Digit 0
  LoadVFDIcon('14SegA',  true,  140,  68, Grid[35], NewGrid[35], 6);
  LoadVFDIcon('14SegB',  true,  156,  71, Grid[34], NewGrid[34], 3);
  LoadVFDIcon('14SegC',  true,  152,  93, Grid[33], NewGrid[33], 1);
  LoadVFDIcon('14SegD',  true,  129, 110, Grid[33], NewGrid[33], 3);
  LoadVFDIcon('14SegE',  true,  126,  93, Grid[33], NewGrid[33], 2);
  LoadVFDIcon('14SegF',  true,  132,  70, Grid[34], NewGrid[34], 2);
  LoadVFDIcon('14SegG1', false, 136,  89, Grid[34], NewGrid[34], 5);
  LoadVFDIcon('14SegG2', false, 147,  89, Grid[34], NewGrid[34], 4);
  LoadVFDIcon('14SegH',  false, 139,  74, Grid[34], NewGrid[34], 0);
  LoadVFDIcon('14SegI',  true,  144,  73, Grid[35], NewGrid[35], 7);
  LoadVFDIcon('14SegJ',  true,  150,  73, Grid[34], NewGrid[34], 1);
  LoadVFDIcon('14SegK',  true,  146,  96, Grid[34], NewGrid[34], 6);
  LoadVFDIcon('14SegL',  true,  138,  93, Grid[33], NewGrid[33], 0);
  LoadVFDIcon('14SegM',  true,  133,  96, Grid[34], NewGrid[34], 7);

  //14 Segment, Digit 1
  LoadVFDIcon('14SegA',  true,  180,  68, Grid[36], NewGrid[36], 0);
  LoadVFDIcon('14SegB',  true,  196,  71, Grid[36], NewGrid[36], 5);
  LoadVFDIcon('14SegC',  true,  192,  93, Grid[35], NewGrid[35], 3);
  LoadVFDIcon('14SegD',  true,  170, 110, Grid[35], NewGrid[35], 5);
  LoadVFDIcon('14SegE',  true,  166,  93, Grid[35], NewGrid[35], 4);
  LoadVFDIcon('14SegF',  true,  172,  70, Grid[36], NewGrid[36], 4);
  LoadVFDIcon('14SegG1', false, 176,  89, Grid[36], NewGrid[36], 7);
  LoadVFDIcon('14SegG2', false, 187,  89, Grid[36], NewGrid[36], 6);
  LoadVFDIcon('14SegH',  false, 179,  74, Grid[36], NewGrid[36], 2);
  LoadVFDIcon('14SegI',  true,  184,  73, Grid[36], NewGrid[36], 1);
  LoadVFDIcon('14SegJ',  true,  190,  74, Grid[36], NewGrid[36], 3);
  LoadVFDIcon('14SegK',  true,  186,  96, Grid[35], NewGrid[35], 0);
  LoadVFDIcon('14SegL',  true,  178,  93, Grid[35], NewGrid[35], 2);
  LoadVFDIcon('14SegM',  true,  173,  96, Grid[35], NewGrid[35], 1);

  //14 Segment, Digit 2
  LoadVFDIcon('14SegA',  true,  222,  68, Grid[38], NewGrid[38], 1);
  LoadVFDIcon('14SegB',  true,  238,  71, Grid[38], NewGrid[38], 6);
  LoadVFDIcon('14SegC',  true,  234,  93, Grid[37], NewGrid[37], 4);
  LoadVFDIcon('14SegD',  true,  211, 110, Grid[37], NewGrid[37], 6);
  LoadVFDIcon('14SegE',  true,  208,  93, Grid[37], NewGrid[37], 5);
  LoadVFDIcon('14SegF',  true,  214,  70, Grid[38], NewGrid[38], 5);
  LoadVFDIcon('14SegG1', false, 218,  89, Grid[37], NewGrid[37], 0);
  LoadVFDIcon('14SegG2', false, 229,  89, Grid[38], NewGrid[38], 7);
  LoadVFDIcon('14SegH',  false, 221,  74, Grid[38], NewGrid[38], 3);
  LoadVFDIcon('14SegI',  true,  226,  73, Grid[38], NewGrid[38], 2);
  LoadVFDIcon('14SegJ',  true,  232,  74, Grid[38], NewGrid[38], 4);
  LoadVFDIcon('14SegK',  true,  229,  96, Grid[37], NewGrid[37], 1);
  LoadVFDIcon('14SegL',  true,  220,  93, Grid[37], NewGrid[37], 3);
  LoadVFDIcon('14SegM',  true,  215,  96, Grid[37], NewGrid[37], 2);

  //14 Segment, Digit 3
  LoadVFDIcon('14SegA',  true,  262,  68, Grid[40], NewGrid[40], 0);
  LoadVFDIcon('14SegB',  true,  278,  71, Grid[40], NewGrid[40], 5);
  LoadVFDIcon('14SegC',  true,  274,  93, Grid[39], NewGrid[39], 3);
  LoadVFDIcon('14SegD',  true,  251, 110, Grid[39], NewGrid[39], 5);
  LoadVFDIcon('14SegE',  true,  248,  93, Grid[39], NewGrid[39], 4);
  LoadVFDIcon('14SegF',  true,  254,  70, Grid[40], NewGrid[40], 4);

  LoadVFDIcon('14SegG1', false, 259,  89, Grid[40], NewGrid[40], 7);
  LoadVFDIcon('14SegG2', false, 270,  89, Grid[40], NewGrid[40], 6);
  LoadVFDIcon('14SegH',  false, 261,  74, Grid[40], NewGrid[40], 2);
  LoadVFDIcon('14SegI',  true,  266,  73, Grid[40], NewGrid[40], 1);
  LoadVFDIcon('14SegJ',  true,  272,  74, Grid[40], NewGrid[40], 3);
  LoadVFDIcon('14SegK',  true,  269,  96, Grid[39], NewGrid[39], 0);
  LoadVFDIcon('14SegL',  true,  260,  93, Grid[39], NewGrid[39], 2);
  LoadVFDIcon('14SegM',  true,  255,  96, Grid[39], NewGrid[39], 1);

  //17 Segment, Digit 0
  LoadVFDIcon('17SegA1', false, 342,  52, Grid[ 5], NewGrid[ 5], 7);   //a1
  LoadVFDIcon('17SegA1', false, 361,  52, Grid[ 5], NewGrid[ 5], 5);   //a2
  LoadVFDIcon('17SegB',  true,  371,  58, Grid[ 3], NewGrid[ 3], 1);   //b
  LoadVFDIcon('17SegB',  true,  365,  87, Grid[ 2], NewGrid[ 2], 7);   //c
  LoadVFDIcon('17SegA1', false, 348, 111, Grid[ 1], NewGrid[ 1], 3);   //d1
  LoadVFDIcon('17SegA1', false, 329, 111, Grid[ 1], NewGrid[ 1], 5);   //d2
  LoadVFDIcon('17SegB',  true,  326,  87, Grid[ 1], NewGrid[ 1], 1);   //e
  LoadVFDIcon('17SegB',  true,  332,  57, Grid[ 4], NewGrid[ 4], 7);   //f
  LoadVFDIcon('17SegG1', false, 335,  81, Grid[ 3], NewGrid[ 3], 7);   //g1
  LoadVFDIcon('17SegG2', false, 351,  81, Grid[ 3], NewGrid[ 3], 3);   //g2
  LoadVFDIcon('17SegG1', false, 358,  81, Grid[ 3], NewGrid[ 3], 5);   //g3
  LoadVFDIcon('17SegI',  true,  352,  58, Grid[ 4], NewGrid[ 4], 1);   //i
  LoadVFDIcon('17SegI',  true,  345,  88, Grid[ 2], NewGrid[ 2], 5);   //l
  LoadVFDIcon('17SegH',  false, 344,  59, Grid[ 4], NewGrid[ 4], 3);   //h
  LoadVFDIcon('17SegJ',  true,  360,  60, Grid[ 4], NewGrid[ 4], 5);   //j
  LoadVFDIcon('17SegH',  false, 355,  91, Grid[ 2], NewGrid[ 2], 1);   //k
  LoadVFDIcon('17SegJ',  true,  334,  93, Grid[ 2], NewGrid[ 2], 3);   //m

  //17 Segment, Digit 1
  LoadVFDIcon('17SegA1', false, 401,  52, Grid[21], NewGrid[21], 7);   //a1
  LoadVFDIcon('17SegA1', false, 420,  52, Grid[21], NewGrid[21], 5);   //a2
  LoadVFDIcon('17SegB',  true,  430,  58, Grid[19], NewGrid[19], 1);   //b
  LoadVFDIcon('17SegB',  true,  424,  87, Grid[18], NewGrid[18], 7);   //c
  LoadVFDIcon('17SegA1', false, 408, 111, Grid[17], NewGrid[17], 3);   //d1
  LoadVFDIcon('17SegA1', false, 389, 111, Grid[17], NewGrid[17], 5);   //d2
  LoadVFDIcon('17SegB',  true,  386,  87, Grid[17], NewGrid[17], 1);   //e
  LoadVFDIcon('17SegB',  true,  393,  58, Grid[20], NewGrid[20], 7);   //f
  LoadVFDIcon('17SegG1', false, 395,  81, Grid[19], NewGrid[19], 7);   //g1
  LoadVFDIcon('17SegG2', false, 411,  81, Grid[19], NewGrid[19], 3);   //g2
  LoadVFDIcon('17SegG1', false, 418,  81, Grid[19], NewGrid[19], 5);   //g3
  LoadVFDIcon('17SegI',  true,  412,  58, Grid[20], NewGrid[20], 1);   //i
  LoadVFDIcon('17SegI',  true,  406,  88, Grid[18], NewGrid[18], 5);   //l
  LoadVFDIcon('17SegH',  false, 404,  59, Grid[20], NewGrid[20], 3);   //h
  LoadVFDIcon('17SegJ',  true,  420,  60, Grid[20], NewGrid[20], 5);   //j
  LoadVFDIcon('17SegH',  false, 416,  91, Grid[18], NewGrid[18], 1);   //k
  LoadVFDIcon('17SegJ',  true,  394,  93, Grid[18], NewGrid[18], 3);   //m

  //17 Segment, Digit 2
  LoadVFDIcon('17SegA1', false, 460,  53, Grid[21], NewGrid[21], 6);   //a1
  LoadVFDIcon('17SegA1', false, 479,  53, Grid[21], NewGrid[21], 4);   //a2
  LoadVFDIcon('17SegB',  true,  490,  58, Grid[19], NewGrid[19], 0);   //b
  LoadVFDIcon('17SegB',  true,  484,  87, Grid[18], NewGrid[18], 6);   //c
  LoadVFDIcon('17SegA1', false, 467, 111, Grid[17], NewGrid[17], 2);   //d1
  LoadVFDIcon('17SegA1', false, 448, 111, Grid[17], NewGrid[17], 4);   //d2
  LoadVFDIcon('17SegB',  true,  445,  87, Grid[17], NewGrid[17], 0);   //e
  LoadVFDIcon('17SegB',  true,  452,  58, Grid[20], NewGrid[20], 6);   //f
  LoadVFDIcon('17SegG1', false, 454,  81, Grid[19], NewGrid[19], 6);   //g1
  LoadVFDIcon('17SegG2', false, 470,  81, Grid[19], NewGrid[19], 2);   //g2
  LoadVFDIcon('17SegG1', false, 478,  81, Grid[19], NewGrid[19], 4);   //g3
  LoadVFDIcon('17SegI',  true,  471,  58, Grid[20], NewGrid[20], 0);   //i
  LoadVFDIcon('17SegI',  true,  464,  88, Grid[18], NewGrid[18], 4);   //l
  LoadVFDIcon('17SegH',  false, 463,  59, Grid[20], NewGrid[20], 2);   //h
  LoadVFDIcon('17SegJ',  true,  479,  60, Grid[20], NewGrid[20], 4);   //j
  LoadVFDIcon('17SegH',  false, 475,  91, Grid[18], NewGrid[18], 0);   //k
  LoadVFDIcon('17SegJ',  true,  454,  93, Grid[18], NewGrid[18], 2);   //m

  //17 Segment, Digit 3
  LoadVFDIcon('17SegA1', false, 521,  53, Grid[ 9], NewGrid[ 9], 5);   //a1
  LoadVFDIcon('17SegA1', false, 540,  53, Grid[ 9], NewGrid[ 9], 3);   //a2
  LoadVFDIcon('17SegB',  true,  550,  59, Grid[ 8], NewGrid[ 8], 7);   //b
  LoadVFDIcon('17SegB',  true,  543,  88, Grid[ 6], NewGrid[ 6], 5);   //c
  LoadVFDIcon('17SegA1', false, 528, 111, Grid[ 5], NewGrid[ 5], 1);   //d1
  LoadVFDIcon('17SegA1', false, 508, 111, Grid[ 5], NewGrid[ 5], 3);   //d2
  LoadVFDIcon('17SegB',  true,  506,  88, Grid[ 6], NewGrid[ 6], 7);   //e
  LoadVFDIcon('17SegB',  true,  512,  58, Grid[ 8], NewGrid[ 8], 5);   //f
  LoadVFDIcon('17SegG1', false, 515,  81, Grid[ 7], NewGrid[ 7], 5);   //g1
  LoadVFDIcon('17SegG2', false, 531,  81, Grid[ 7], NewGrid[ 7], 1);   //g2
  LoadVFDIcon('17SegG1', false, 538,  81, Grid[ 7], NewGrid[ 7], 3);   //g3
  LoadVFDIcon('17SegI',  true,  531,  59, Grid[ 9], NewGrid[ 9], 7);   //i
  LoadVFDIcon('17SegI',  true,  525,  88, Grid[ 6], NewGrid[ 6], 3);   //l
  LoadVFDIcon('17SegH',  false, 523,  59, Grid[ 8], NewGrid[ 8], 1);   //h
  LoadVFDIcon('17SegJ',  true,  539,  60, Grid[ 8], NewGrid[ 8], 3);   //j
  LoadVFDIcon('17SegH',  false, 535,  92, Grid[ 7], NewGrid[ 7], 7);   //k
  LoadVFDIcon('17SegJ',  true,  514,  93, Grid[ 6], NewGrid[ 6], 1);   //m

  //17 Segment, Digit 4
  LoadVFDIcon('17SegA1', false, 580,  53, Grid[ 9], NewGrid[ 9], 4);   //a1
  LoadVFDIcon('17SegA1', false, 599,  53, Grid[ 9], NewGrid[ 9], 2);   //a2
  LoadVFDIcon('17SegB',  true,  609,  59, Grid[ 8], NewGrid[ 8], 6);   //b
  LoadVFDIcon('17SegB',  true,  603,  88, Grid[ 6], NewGrid[ 6], 4);   //c
  LoadVFDIcon('17SegA1', false, 587, 111, Grid[ 5], NewGrid[ 5], 0);   //d1
  LoadVFDIcon('17SegA1', false, 568, 111, Grid[ 5], NewGrid[ 5], 2);   //d2
  LoadVFDIcon('17SegB',  true,  565,  88, Grid[ 6], NewGrid[ 6], 6);   //e
  LoadVFDIcon('17SegB',  true,  571,  58, Grid[ 8], NewGrid[ 8], 4);   //f
  LoadVFDIcon('17SegG1', false, 574,  82, Grid[ 7], NewGrid[ 7], 4);   //g1
  LoadVFDIcon('17SegG2', false, 590,  82, Grid[ 7], NewGrid[ 7], 0);   //g2
  LoadVFDIcon('17SegG1', false, 597,  82, Grid[ 7], NewGrid[ 7], 2);   //g3
  LoadVFDIcon('17SegI',  true,  590,  59, Grid[ 9], NewGrid[ 9], 6);   //i
  LoadVFDIcon('17SegI',  true,  585,  88, Grid[ 6], NewGrid[ 6], 2);   //l
  LoadVFDIcon('17SegH',  false, 582,  59, Grid[ 8], NewGrid[ 8], 0);   //h
  LoadVFDIcon('17SegJ',  true,  599,  60, Grid[ 8], NewGrid[ 8], 2);   //j
  LoadVFDIcon('17SegH',  false, 594,  92, Grid[ 7], NewGrid[ 7], 6);   //k
  LoadVFDIcon('17SegJ',  true,  573,  93, Grid[ 6], NewGrid[ 6], 0);   //m

  //17 Segment, Digit 5
  LoadVFDIcon('17SegA1', false, 640,  53, Grid[25], NewGrid[25], 5);   //a1
  LoadVFDIcon('17SegA1', false, 660,  53, Grid[25], NewGrid[25], 3);   //a2
  LoadVFDIcon('17SegB',  true,  670,  59, Grid[24], NewGrid[24], 7);   //b
  LoadVFDIcon('17SegB',  true,  664,  87, Grid[22], NewGrid[22], 5);   //c
  LoadVFDIcon('17SegA1', false, 647, 112, Grid[21], NewGrid[21], 1);   //d1
  LoadVFDIcon('17SegA1', false, 628, 112, Grid[21], NewGrid[21], 3);   //d2
  LoadVFDIcon('17SegB',  true,  626,  88, Grid[22], NewGrid[22], 7);   //e
  LoadVFDIcon('17SegB',  true,  631,  59, Grid[24], NewGrid[24], 5);   //f
  LoadVFDIcon('17SegG1', false, 634,  82, Grid[23], NewGrid[23], 5);   //g1
  LoadVFDIcon('17SegG2', false, 650,  82, Grid[23], NewGrid[23], 1);   //g2
  LoadVFDIcon('17SegG1', false, 657,  82, Grid[23], NewGrid[23], 3);   //g3
  LoadVFDIcon('17SegI',  true,  651,  59, Grid[25], NewGrid[25], 7);   //i
  LoadVFDIcon('17SegI',  true,  645,  88, Grid[22], NewGrid[22], 3);   //l
  LoadVFDIcon('17SegH',  false, 643,  60, Grid[24], NewGrid[24], 1);   //h
  LoadVFDIcon('17SegJ',  true,  659,  61, Grid[24], NewGrid[24], 3);   //j
  LoadVFDIcon('17SegH',  false, 655,  92, Grid[23], NewGrid[23], 7);   //k
  LoadVFDIcon('17SegJ',  true,  633,  94, Grid[22], NewGrid[22], 1);   //m

  //17 Segment, Digit 6
  LoadVFDIcon('17SegA1', false, 700,  53, Grid[25], NewGrid[25], 4);   //a1
  LoadVFDIcon('17SegA1', false, 719,  53, Grid[25], NewGrid[25], 2);   //a2
  LoadVFDIcon('17SegB',  true,  729,  59, Grid[24], NewGrid[24], 6);   //b
  LoadVFDIcon('17SegB',  true,  723,  88, Grid[22], NewGrid[22], 4);   //c
  LoadVFDIcon('17SegA1', false, 707, 112, Grid[21], NewGrid[21], 0);   //d1
  LoadVFDIcon('17SegA1', false, 688, 112, Grid[21], NewGrid[21], 2);   //d2
  LoadVFDIcon('17SegB',  true,  685,  87, Grid[22], NewGrid[22], 6);   //e
  LoadVFDIcon('17SegB',  true,  691,  59, Grid[24], NewGrid[24], 4);   //f
  LoadVFDIcon('17SegG1', false, 694,  82, Grid[23], NewGrid[23], 4);   //g1
  LoadVFDIcon('17SegG2', false, 710,  82, Grid[23], NewGrid[23], 0);   //g2
  LoadVFDIcon('17SegG1', false, 717,  82, Grid[23], NewGrid[23], 2);   //g3
  LoadVFDIcon('17SegI',  true,  710,  59, Grid[25], NewGrid[25], 6);   //i
  LoadVFDIcon('17SegI',  true,  704,  89, Grid[22], NewGrid[22], 2);   //l
  LoadVFDIcon('17SegH',  false, 702,  60, Grid[24], NewGrid[24], 0);   //h
  LoadVFDIcon('17SegJ',  true,  719,  61, Grid[24], NewGrid[24], 2);   //j
  LoadVFDIcon('17SegH',  false, 714,  92, Grid[23], NewGrid[23], 6);   //k
  LoadVFDIcon('17SegJ',  true,  693,  94, Grid[22], NewGrid[22], 0);   //m

  //17 Segment, Digit 7
  LoadVFDIcon('17SegA1', false, 760,  53, Grid[ 5], NewGrid[ 5], 6);   //a1
  LoadVFDIcon('17SegA1', false, 779,  53, Grid[ 5], NewGrid[ 5], 4);   //a2
  LoadVFDIcon('17SegB',  true,  790,  59, Grid[ 3], NewGrid[ 3], 0);   //b
  LoadVFDIcon('17SegB',  true,  784,  88, Grid[ 2], NewGrid[ 2], 6);   //c
  LoadVFDIcon('17SegA1', false, 768, 112, Grid[ 1], NewGrid[ 1], 2);   //d1
  LoadVFDIcon('17SegA1', false, 749, 112, Grid[ 1], NewGrid[ 1], 4);   //d2
  LoadVFDIcon('17SegB',  true,  746,  88, Grid[ 1], NewGrid[ 1], 0);   //e
  LoadVFDIcon('17SegB',  true,  751,  59, Grid[ 4], NewGrid[ 4], 6);   //f
  LoadVFDIcon('17SegG1', false, 754,  82, Grid[ 3], NewGrid[ 3], 6);   //g1
  LoadVFDIcon('17SegG2', false, 770,  82, Grid[ 3], NewGrid[ 3], 2);   //g2
  LoadVFDIcon('17SegG1', false, 777,  82, Grid[ 3], NewGrid[ 3], 4);   //g3
  LoadVFDIcon('17SegI',  true,  771,  59, Grid[ 4], NewGrid[ 4], 0);   //i
  LoadVFDIcon('17SegI',  true,  765,  88, Grid[ 2], NewGrid[ 2], 4);   //l
  LoadVFDIcon('17SegH',  false, 762,  60, Grid[ 4], NewGrid[ 4], 2);   //h
  LoadVFDIcon('17SegJ',  true,  779,  61, Grid[ 4], NewGrid[ 4], 4);   //j
  LoadVFDIcon('17SegH',  false, 775,  92, Grid[ 2], NewGrid[ 2], 0);   //k
  LoadVFDIcon('17SegJ',  true,  754,  94, Grid[ 2], NewGrid[ 2], 2);   //m

  LoadVFDIcon('CDA',     true,    8,  72, Grid[25], NewGrid[25], 0);
  LoadVFDIcon('CDB',     true,    8,  72, Grid[26], NewGrid[26], 7);
  LoadVFDIcon('CDC',     true,    8,  72, Grid[26], NewGrid[26], 6);
  LoadVFDIcon('CDD',     true,    8,  72, Grid[26], NewGrid[26], 5);
  LoadVFDIcon('CDE',     true,    8,  72, Grid[26], NewGrid[26], 4);
  LoadVFDIcon('CDF',     true,    8,  72, Grid[26], NewGrid[26], 3);
  LoadVFDIcon('CDG',     true,    8,  72, Grid[26], NewGrid[26], 2);
  LoadVFDIcon('CDH',     true,    8,  72, Grid[26], NewGrid[26], 1);
  LoadVFDIcon('CDI',     true,    8,  72, Grid[26], NewGrid[26], 0);
  LoadVFDIcon('CDJ',     true,    8,  72, Grid[27], NewGrid[27], 7);
  LoadVFDIcon('CDK',     true,    8,  72, Grid[27], NewGrid[27], 6);
  LoadVFDIcon('CDL',     true,    8,  72, Grid[27], NewGrid[27], 5);
  LoadVFDIcon('CDCenter',true,    8,  72, Grid[25], NewGrid[25], 1);

//  SourceRect.TopLeft := Point(0, 0);
//  SourceRect.BottomRight := Point(Image_VFD.Width, Image_VFD.Height);
//  DestRect.TopLeft := Point(0, 0);
//  DestRect.BottomRight := Point(Image_Display.Width, Image_Display.Height);
//  Image_Display.Canvas.CopyRect(DestRect, Image_VFD.Canvas, SourceRect);

  copymemory(@Grid[0], NewGrid, 48);
end;
{$ENDREGION}

{$REGION 'Network'}
procedure TForm1.CommOpen();

var
  CP                   : ^tConnectionProfile;

begin
  CP := @ConnectionProfile[CurrentConnectionProfile];

  WSocket1.Addr := CP.ToppyIP;
  WSocket1.Port := format('%d', [CP.ToppyVFDPort]);
  WSocket1.Proto := 'tcp';

  try
    WSocket1.Connect;

  except

  end;
end;

procedure TForm1.CommClose();
begin
  WSocket1.Close();
end;

procedure TForm1.SendAlive();

var
  Cmd                  : char;

begin
  if CommConnected and (GetTickCount() > AliveTimer)then
  begin
    AliveTimer := GetTickCount() + 1000;


    Cmd := #13;
    WSocket1.Send(@Cmd, 1);
  end;
end;

procedure TForm1.WSocket1DataAvailable(Sender: TObject; ErrCode: Word);

var
  Socket               : TWSocket;
  Buffer               : array[0..47] of byte;

begin
  if ErrCode <> 0 then exit;

  Socket := Sender as TWSocket;

  while Socket.RcvdCount > 0 do
  begin
    outputdebugstring(pchar(format('Rcv=%d', [Socket.RcvdCount])));
//    if Socket.RcvdCount = 48 then
//    begin
      Socket.Receive(@Buffer[0], 48);
      UpdateVFD(@Buffer[0]);
//    end;
  end;
end;

procedure TForm1.WSocket1SessionConnected(Sender: TObject; ErrCode: Word);
begin
  CommConnected := true;
  WSocket1.SocketRcvBufSize := $20000;
  Button_Connect.Caption := 'Disconnect';
  Button_Connect.Enabled := true;
end;

procedure TForm1.WSocket1SessionClosed(Sender: TObject; ErrCode: Word);
begin
  CommConnected := false;
  Button_Connect.Caption := 'Connect';
  Button_Connect.Enabled := true;
  ClearVFD();
end;
{$ENDREGION}

end.

