#include                <stdlib.h>
#include                <string.h>
#include                <sys/socket.h>
#include                <arpa/inet.h>
#include                <unistd.h>
#include                <netinet/in.h>
#include                <fcntl.h>
#include                <errno.h>
#include                "NetworkVFD.h"
#include                "TMSRemote.h"
#include                "libFireBird.h"

#ifndef ROODIR
  #define ROOTDIR       "/ProgramFiles/Settings/" PROGRAM_NAME
#endif

int                     Settings_MaxPendingVFD;
int                     Settings_BufferSizeVFD;
int                     Settings_PortVFD;
dword                   AliveTimerVFD = 0;

extern void WriteLog(char *s);

char                    LogPuffer[256];
#define DEBUG(...)      {sprintf(LogPuffer, __VA_ARGS__); WriteLog(LogPuffer);}

//-------------------------------------------------------------------------------------------------------
// Settings
//-------------------------------------------------------------------------------------------------------
void CreateRootDirVFD(void)
{
  //Check & Create Folders
  HDD_TAP_PushDir();
  HDD_ChangeDir("/ProgramFiles");
  if(!TAP_Hdd_Exist("Settings")) TAP_Hdd_Create("Settings", ATTR_FOLDER);
  HDD_ChangeDir("Settings");
  if(!TAP_Hdd_Exist(PROGRAM_NAME)) TAP_Hdd_Create(PROGRAM_NAME, ATTR_FOLDER);
  HDD_TAP_PopDir();
}

void LoadSettingsVFD()
{
  INILOCATION           ProfileLocation;

  CreateRootDirVFD();
  TAP_Hdd_ChangeDir(ROOTDIR);
  ProfileLocation = INIOpenFile("TMSRemote_Network.ini", NULL);

  if(ProfileLocation > INILOCATION_NewFile)
  {
    Settings_MaxPendingVFD = INIGetInt ("MaxPending", 1, 1, 1);
    Settings_BufferSizeVFD = INIGetInt ("BufferSize", 65535,  1, 65535);
    Settings_PortVFD = INIGetInt ("PortVFD", NETWORKPORTVFD,  1, 65535);
    INICloseFile();
  }
  else  //load defaults
  {
    DEBUG("VFD: %s not found, using default settings", "TMSRemote_Network.ini");
    Settings_MaxPendingVFD    = 1;
    Settings_BufferSizeVFD    = 65535;
    Settings_PortVFD          = NETWORKPORTVFD;
  }
  DEBUG("VFD: Listening on TCP port %d", Settings_PortVFD);
}

//-------------------------------------------------------------------------------------------------------
// Socket functions
//-------------------------------------------------------------------------------------------------------

static int                    ServerSockVFD = 0;
static int                    ClientSockVFD = 0;
static struct sockaddr_in     echoclientVFD;

void SetupSocketServerVFD()
{
  struct sockaddr_in    echoserver;
  int                   flags;

  // Create the TCP socket
  if ((ServerSockVFD = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
  {
    DEBUG("VFD: Failed to create socket");
    return;
  }
  //DEBUG("VFD: Server socket ID = 0x%8.8x", ServerSock);

  //set non blocking mode
  flags = fcntl(ServerSockVFD, F_GETFL, 0);
  fcntl(ServerSockVFD, F_SETFL, flags | O_NONBLOCK);

  // Construct the server sockaddr_in structure
  memset(&echoserver, 0, sizeof(echoserver));       // Clear struct
  echoserver.sin_family = AF_INET;                  // Internet/IP
  echoserver.sin_addr.s_addr = htonl(INADDR_ANY);   // Incoming addr
  echoserver.sin_port = htons(Settings_PortVFD);    // server port

  //DEBUG("VFD: Binding server socket on port %d...", Settings_Port);
  // Bind the server socket
  if (bind(ServerSockVFD, (struct sockaddr *) &echoserver, sizeof(echoserver)) < 0)
  {
    DEBUG("VFD: Failed to bind the server socket");
    return;
  }

  //DEBUG("VFD: Listening on server socket...");
  // Listen on the server socket
  if (listen(ServerSockVFD, Settings_MaxPendingVFD) < 0)
  {
    DEBUG("VFD: Failed to listen on server socket");
    return;
  }
}

bool CloseSocketVFD(void)
{
  if(ClientSockVFD > 0)
  {
    DEBUG("VFD: Closing connection to client");
    close(ClientSockVFD);
    ClientSockVFD = 0;
    AliveTimerVFD = 0;

    return TRUE;
  }

  return FALSE;
}

void CheckSocketVFD(void)
{
  int                   received = -1;
  unsigned int          clientlen = sizeof(echoclientVFD);
  char                  Buffer[100];

  if(!ServerSockVFD) return;

  // Wait for client connection
  if(ClientSockVFD <= 0)
  {
    if ((ClientSockVFD = accept(ServerSockVFD, (struct sockaddr *) &echoclientVFD, &clientlen)) < 0)
    {
      if(errno == EAGAIN) return;   //nothing waiting

      DEBUG("VFD: Failed to accept client connection, error = %d", ClientSockVFD);
      return;
    }
    DEBUG("VFD: Client connected: ID = 0x%8.8x, %s", ClientSockVFD, inet_ntoa(echoclientVFD.sin_addr));
    AliveTimerVFD = TAP_GetTick() + ALIVETIMEOUT;
  }

  // Receive message
  if ((ClientSockVFD > 0) && ((received = recv(ClientSockVFD, Buffer, 100, MSG_DONTWAIT)) < 0))
  {
    if(errno == EAGAIN) return;   //nothing waiting

    DEBUG("VFD: Failed to receive bytes from client, error = 0x%8.8x", errno);
    return;
  }

  if(received == 0)
    CloseSocketVFD();
  else
    AliveTimerVFD = TAP_GetTick() + ALIVETIMEOUT;
}

bool SendSocketVFD(byte *Buffer, int BufferLen)
{
  if(ServerSockVFD && (ClientSockVFD > 0))
  {
    if(send(ClientSockVFD, Buffer, BufferLen, 0) != BufferLen)
    {
      DEBUG("VFD: Failed to send bytes to client");
      return FALSE;
    }
    return TRUE;
  }

  return FALSE;
}

void ShutdownSocketServerVFD()
{
  if(ServerSockVFD)
  {
    DEBUG("VFD: Closing server socket");
    close(ServerSockVFD);
    ServerSockVFD = 0;
  }
}

bool ActiveConnectionVFD(void)
{
  return ((ServerSockVFD > 0) && (ClientSockVFD > 0));
}

void VFDSendGrid(void)
{
  static byte           *grid = NULL;
  static byte            LastGrid[48];

  if(!grid)
  {
    grid = (byte*)FIS_vGrid();
    if(!grid) return;
  }

  if(NetworkVFD_IsAnyConnection())
  {
    if(memcmp(LastGrid, grid, 48))
    {
      memcpy(LastGrid, grid, 48);
      SendSocketVFD(LastGrid, 48);
    }
  }
  else
    LastGrid[0] = 255;
}


//-------------------------------------------------------------------------------------------------------
// State machine
//-------------------------------------------------------------------------------------------------------
bool NetworkVFD_IsAnyConnection(void)
{
  return (ClientSockVFD > 0);
}

void NetworkVFD_Close(void)
{
  DEBUG("VFD: Shutdown event");
  CloseSocketVFD();
  ShutdownSocketServerVFD();
}

dword NetworkVFD_EventHandler(word event, dword param1, dword param2)
{
  static bool           doNotReenter = FALSE;
  static bool           Init = TRUE;
  static bool           Exit = FALSE;

  UNUSED(param2);

  //Don't do anything if the network server has been closed
  if(Exit) return param1;

  if(doNotReenter) return param1;

  doNotReenter = TRUE;

  //First initialization
  if(Init)
  {
    LoadSettingsVFD();
    SetupSocketServerVFD();
    Init = FALSE;
  }

  //An event controlled shutdown
  if(event == EVT_STOP && param1 < 2)
  {
    Exit = TRUE;
    NetworkVFD_Close();
    doNotReenter = FALSE;
    return param1;
  }

  //Manage the network interface
  if(event == EVT_IDLE)
  {
    CheckSocketVFD();
    VFDSendGrid();

    if(AliveTimerVFD && (TAP_GetTick() > AliveTimerVFD))
    {
      DEBUG("VFD: VFD ALIVE timeout reached");
      CloseSocketVFD();
    }
  }

  doNotReenter = FALSE;
  return param1;
}
