#include                <string.h>
#include                "tap.h"
#include                "libFireBird.h"
#include                "TMSRemote.h"
#include                "Network.h"
#include                "NetworkVFD.h"
#include                "TMSCommander.h"
#include                "TMSRemote_TAPCOM.h"

TAP_ID                  (0x8E0A4230);
TAP_PROGRAM_NAME        (PROGRAM_NAME" "VERSION);
TAP_PROGRAM_VERSION     (VERSION);
TAP_AUTHOR_NAME         ("FireBird/jkIT");
TAP_DESCRIPTION         ("TMSRemote");
TAP_ETCINFO             (__DATE__);

byte                    MacroblockChanged[1620];
dword                  *OSDBackup = NULL;
bool                    CaptureVideo = FALSE;
TYPE_VideoFrame         VideoFrameMain;
dword                   VideoNextFrame = 0;
extern bool             HighSpeedUpdate;

enum
{
  MB_TAP     = 1 << 0,
  MB_SUBT    = 1 << 1,
  MB_BASE    = 1 << 2,
  MB_BACKUP  = 1 << 3,
  MB_CHANGED = 1 << 4
};

void WriteLog(char *s)
{
  static bool           FirstCall = TRUE;

  HDD_TAP_PushDir();
  if(FirstCall)
  {
    TAP_Hdd_ChangeDir("/ProgramFiles");
    if(!TAP_Hdd_Exist("Settings")) TAP_Hdd_Create("Settings", ATTR_FOLDER);
    TAP_Hdd_ChangeDir("Settings");
    if(!TAP_Hdd_Exist(PROGRAM_NAME)) TAP_Hdd_Create(PROGRAM_NAME, ATTR_FOLDER);
    FirstCall = FALSE;
  }
  TAP_Hdd_ChangeDir(LOGDIR);
  LogEntry(LOGFILE, PROGRAM_NAME, TRUE, TIMESTAMP_YMDHMS, s);
  HDD_TAP_PopDir();
}

void ResetMacroblockChanged(void)
{
  int                   i;

  for(i = 0; i < 1620; i++)
    MacroblockChanged[i] = 0;
}

void SendMacroblock(dword MacroblockIndex, dword MacroblockOffset, void *StartOfBuffer)
{
  tMacroBlock           MacroBlock;
  register int          y, x;
  int                   y0, x0, RLCnt;
  dword                *pOrig;
  dword                 Pixel, LastPixel = 0xfefefefe;

  y0 = (int)(MacroblockIndex / 45) << 4;
  x0 = (int)(MacroblockIndex % 45) << 4;

  MacroBlock.BlockIndex = MacroblockIndex + MacroblockOffset;
  MacroBlock.RLELength  = 0;
  RLCnt = 0;

  for(y = 0; y < 16; y++)
  {
    pOrig = (dword*)(StartOfBuffer + ((y0 + y) * 720 + x0) * 4);
    for(x = 0; x < 16; x++)
    {
      Pixel = *pOrig & 0x00ffffff;

      //0 is reserved for background pixels. Don't allow them here.
      if(Pixel == 0) Pixel = 0x00000001;

      if(Pixel == LastPixel)
      {
        RLCnt++;
      }
      else
      {
        if(MacroBlock.RLELength > 0) MacroBlock.RLEData[MacroBlock.RLELength - 1] |= (RLCnt << 24);

        RLCnt = 0;
        MacroBlock.RLEData[MacroBlock.RLELength] = Pixel;
        MacroBlock.RLELength++;
        LastPixel = Pixel;
      }

      pOrig++;
    }
  }
  if(MacroBlock.RLELength > 0) MacroBlock.RLEData[MacroBlock.RLELength - 1] |= (RLCnt << 24);

  SendSocket((byte*)&MacroBlock, 8 + (MacroBlock.RLELength << 2));
}

void VideoCapture(void)
{
  bool                  VideoCaptureOK;
  dword                *pVideo;
  register int          y, x, p, MacroblockIndex;

  if(VideoFrameMain.data)
  {
    TAP_MemFree(VideoFrameMain.data);
    VideoFrameMain.data = NULL;
  }

  VideoCaptureOK = TAP_CaptureScreen(CHANNEL_Main, &VideoFrameMain, 0) == 0;
  if(VideoCaptureOK)
  {
    for(y = 0; y < 576; y += 16)
    {
      p = (y >> 4) * 45;
      pVideo = (dword*)(VideoFrameMain.data + y * 720 * 4);

      for(x = 0; x < 720; x += 16)
      {
        MacroblockIndex = p + (x >> 4);
        //TAP_PrintNet("Vid   MB %d\n", MacroblockIndex);
        SendMacroblock(MacroblockIndex, 3000, VideoFrameMain.data);
        pVideo++;
      }
    }
  }
}

void OSDDump(void)
{
  TYPE_OsdBaseInfo      OsdBaseInfoBase, OsdBaseInfoTAP, OsdBaseInfoSubt;
  register int          y, x, p, MacroblockIndex;
  dword                *pOrigBase, *pOrigTAP, *pOrigSubt, *pBackup;
  tMacroBlock           MacroBlock;
  bool					        osdChanged = FALSE;

  ResetMacroblockChanged();
  TAP_Osd_GetPlaneBaseInfo(&OsdBaseInfoBase, BASE_PLANE);
  TAP_Osd_GetPlaneBaseInfo(&OsdBaseInfoTAP, SUBT_PLANE);   //seems to be TAP_PLANE
  TAP_Osd_GetPlaneBaseInfo(&OsdBaseInfoSubt, TAP_PLANE);   //seems to be TAP_PLANE

  for(y = 0; y < 576; y++)
  {
    p = (y >> 4) * 45;

    pOrigBase = (dword*)(OsdBaseInfoBase.frameBuffer + y * 720 * 4);
    pOrigTAP  = (dword*)(OsdBaseInfoTAP.frameBuffer + y * 720 * 4);
    pOrigSubt = (dword*)(OsdBaseInfoSubt.frameBuffer + y * 720 * 4);
    pBackup = (dword*)(OSDBackup + y * 720);

    for(x = 0; x < 720; x++)
    {
      MacroblockIndex = p + (x >> 4);

      if(*pOrigTAP != 0)
      {
        MacroblockChanged[MacroblockIndex] |= MB_TAP;

        if(*pOrigTAP != *pBackup)
        {
          *pBackup = *pOrigTAP;
          MacroblockChanged[MacroblockIndex] |= MB_CHANGED;
        }
      }
      else if(*pOrigSubt != 0)
      {
        MacroblockChanged[MacroblockIndex] |= MB_SUBT;

        if(*pOrigSubt != *pBackup)
        {
          *pBackup = *pOrigSubt;
          MacroblockChanged[MacroblockIndex] |= MB_CHANGED;
        }
      }
      else if(*pOrigBase != 0)
      {
        MacroblockChanged[MacroblockIndex] |= MB_BASE;

        if(*pOrigBase != *pBackup)
        {
          *pBackup = *pOrigBase;
          MacroblockChanged[MacroblockIndex] |= MB_CHANGED;
        }
      }
      else
      {
        if(0 != *pBackup)
        {
          *pBackup = 0;
          MacroblockChanged[MacroblockIndex] |= MB_BACKUP;
        }
      }
      pOrigBase++;
      pOrigTAP++;
      pOrigSubt++;
      pBackup++;
    }
  }

  for (p = 0; p < 1620; p++)
  {
    if (MacroblockChanged[p] & MB_CHANGED || (MacroblockChanged[p] & MB_BACKUP && MacroblockChanged[p] & (MB_TAP | MB_SUBT | MB_BASE)))
    {
      osdChanged = TRUE;

      if (MacroblockChanged[p] & MB_TAP) SendMacroblock(p, 0, OsdBaseInfoTAP.frameBuffer);
      else if (MacroblockChanged[p] & MB_SUBT) SendMacroblock(p, 0, OsdBaseInfoSubt.frameBuffer);
      else if (MacroblockChanged[p] & MB_BASE) SendMacroblock(p, 0, OsdBaseInfoBase.frameBuffer);
    }
    else if (MacroblockChanged[p] & MB_BACKUP)
    {
      osdChanged = TRUE;

      MacroBlock.BlockIndex = p;
      MacroBlock.RLELength  = 1;
      MacroBlock.RLEData[0] = 0xff000000;
      SendSocket((byte*) &MacroBlock, 12);
    }
  }

  // send 'frame complete' message, if OSD has changed
  if(osdChanged) SendMessage(MID_OSDframeComplete);
}

bool isAnyMacroblockChanged(void)
{
  register int          i;

  for(i = 0; i < 1620; i++)
    if(MacroblockChanged[i]) return TRUE;

  return FALSE;
}

void ExitTAP(void)
{
  Network_Close();
  if(OSDBackup)
  {
    TAP_MemFree(OSDBackup);
    OSDBackup = NULL;
  }

  if(VideoFrameMain.data)
  {
    TAP_MemFree(VideoFrameMain.data);
    VideoFrameMain.data = NULL;
  }
  TAP_Exit();
}

dword TMSCommander_handler(dword param1)
{
  switch (param1)
  {
    case TMSCMDR_Capabilities:
      return (dword)(TMSCMDR_CanBeStopped);

    case TMSCMDR_Menu:
      return TMSCMDR_NotOK;

    case TMSCMDR_Stop:
      ExitTAP();
      return TMSCMDR_OK;

    default:
      return TMSCMDR_UnknownFunction;
  }
}

dword TAP_EventHandler(word event, dword param1, dword param2)
{
  static bool           Init = TRUE;
  static bool           LastNextConnection = FALSE;
  static dword          NextOSDCheck = 0;

  (void) event;
  (void) param2;

  Network_EventHandler(event, param1, param2);
  NetworkVFD_EventHandler(event, param1, param2);

  if(Init)
  {
    memset(&VideoFrameMain, 0, sizeof(VideoFrameMain));
    VideoFrameMain.pixelFormat = OSD_8888;
    VideoFrameMain.width = 720;
    VideoFrameMain.height = 576;

    Init = FALSE;
  }

  if(event == EVT_IDLE)
  {
    if(TAP_GetTick() > NextOSDCheck)
    {
      if(Network_IsAnyConnection())
      {
        if(!LastNextConnection) memset(OSDBackup, 0xfe, 720*576*4);
        OSDDump();
      }
      LastNextConnection = Network_IsAnyConnection();
      if(HighSpeedUpdate)
        NextOSDCheck = TAP_GetTick() + 10;
      else
        NextOSDCheck = TAP_GetTick() + 50;
    }
  }

  if(event == EVT_STOP && param1 < 2)
  {
    ExitTAP();
  }

  if (event == EVT_TMSCommander)
  {
    return TMSCommander_handler(param1);
  }

  if (event == EVT_TAPCOM)
  {
    TAPCOM_Channel  Channel;
    dword           ServiceID;

    Channel = TAPCOM_GetChannel(param1, NULL, &ServiceID, NULL, NULL);
    switch(ServiceID)
    {
      case TMSREMOTE_TAPCOM_SERVICE_STATUS:
      {
        int ret;

        ret = (Network_IsOsdHidden() ? 2 : 0) | (Network_IsAnyConnection() ? 1 : 0);
        TAPCOM_Finish(Channel, ret);
        break;
      }

      case TMSREMOTE_TAPCOM_SERVICE_SETDIRECT:
      {
        Network_SendDirectOn();
        TAPCOM_Finish(Channel, 0);
        break;
      }

      case TMSREMOTE_TAPCOM_SERVICE_CLEARDIRECT:
      {
        Network_SendDirectOff();
        TAPCOM_Finish(Channel, 0);
        break;
      }

      default:
        TAPCOM_Reject(Channel);
        break;
    }
  }

  return param1;
}

int TAP_Main(void)
{
  OSDBackup = TAP_MemAlloc(720*576*4);

  return 1;
}
