/*
 * websocket.c
 *
 *  Created on: 12.09.2013
 *      Author: jkIT
 */

#include 				<stdio.h>
#include 				<stdlib.h>
#include 				<string.h>
#include 				<time.h>
#include                "websocket.h"
#include                "tap_debug.h"
#include                "sha1.h"

/**
 * base64 Implementation based on
 * http://base64.sourceforge.net/
 */

/**
 * Translation Table for base64encoding as described in RFC1113
 */
static const char cb64[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

/**
 * (base64) encode 3 8-bit binary bytes as 4 '6-bit' characters
 */
void b64encodeblock( unsigned char *in, char *out, int len )
{
    out[0] = (char) cb64[ (int)(in[0] >> 2) ];
    out[1] = (char) cb64[ (int)(((in[0] & 0x03) << 4) | ((in[1] & 0xf0) >> 4)) ];
    out[2] = (char) (len > 1 ? cb64[ (int)(((in[1] & 0x0f) << 2) | ((in[2] & 0xc0) >> 6)) ] : '=');
    out[3] = (char) (len > 2 ? cb64[ (int)(in[2] & 0x3f) ] : '=');
}

/**
 * (base64) encode 'len' 8-bit binary bytes as string
 */
int b64encode( unsigned char *in, char *out, int len ) {

	unsigned char *_in = in;
	char *_out = out;
	int i;
	for (i = 0; i < len; i += 3, _in += 3, _out += 4) {
		b64encodeblock(_in, _out, (len - i >= 3) ? 3 : len - i);
	}

	_out[0] = 0;

	return out - _out;

}

/**
 * parse HTTP header (connection request)
 *
 * returns 0 on failure, else length
 */
int ws_parse_header(char *s, int maxlen, tRequestHeader *header) {

	DEBUG("ws_parse_header: (length: %d)", maxlen);

	header->host[0] = 0;
	header->origin[0] = 0;
	header->sec_key[0] = 0;
	header->protocol[0] = 0;
	header->ws_version = 0;

	s[maxlen] = '\0';
	char *ptr = s;

	ptr = strtok(s, "\r\n");

	DEBUG("| %s", ptr);

	// first header line should be 'GET /tmsremote HTTP/1.1'
	if ( strcmp(ptr, "GET /tmsremote HTTP/1.1") != 0) {

		DEBUG("!!! invalid request: %s", ptr);
		return 0;

	}

	// rest of the header
	ptr = strtok(NULL, "\r\n");
	while(ptr != NULL) {

		DEBUG("| %s", ptr);

		if (
				( strncasecmp( ptr, "Sec-WebSocket-Version: ", 23 ) == 0 ) &&
				( strlen( ptr ) > 23)
		) {

			header->ws_version = strtol( ptr + 23, NULL, 10 );

		} else if (
				( strncasecmp( ptr, "Sec-WebSocket-Key: ", 19 ) == 0 ) &&
				( strlen( ptr ) > 19 )
		) {

			strcpy( header->sec_key, ptr + 19 );

			char sha1code[] = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11";

			sha1_context ctx;
		    unsigned char sha1sum[21];

			sha1_starts( &ctx );
			sha1_update( &ctx, (uint8 *) &header->sec_key, strlen( (char*) &header->sec_key ) );
			sha1_update( &ctx, (uint8 *) sha1code, strlen( sha1code ) );
			sha1_finish( &ctx, sha1sum);
			sha1sum[20] = 0;

			b64encode( sha1sum, header->sec_accept, 20 );

		} else if (
				( strncasecmp( ptr, "Sec-WebSocket-Origin: ", 22 ) == 0 ) &&
				( strlen( ptr) > 22 )
		) {

			strcpy( header->origin, ptr + 22 );

		} else if (
				( strncasecmp(ptr, "Origin: ", 8 ) == 0 ) &&
				( strlen( ptr ) > 8 )
		) {

			strcpy( header->origin, ptr + 8 );

		} else if (
				( strncasecmp( ptr, "Sec-WebSocket-Protocol: ", 24 ) == 0 ) &&
				( strlen( ptr ) > 24)
		) {

			strcpy( header->protocol, ptr + 24 );

		} else if (
				( strncasecmp( ptr, "Host: ", 6 ) == 0 ) &&
				( strlen( ptr ) > 6)
		) {

			strcpy( header->host, ptr + 6 );

		}

		ptr = strtok( NULL, "\r\n" );

	}

	if ( header->ws_version < 6 ) {

		DEBUG( "ws_header: invalid websocket version: %d (<6)", header->ws_version );
		return 0;

	}

	if ( strlen( header->origin ) == 0 ) {
		DEBUG( "ws_header: Origin is missing" );
		return 0;
	}

	DEBUG("found:");
	DEBUG("- Host=%s", header->host);
	DEBUG("- Origin=%s", header->origin);
	DEBUG("- Sec-WebSocket-Version=%d", header->ws_version);
	DEBUG("- Sec-WebSocket-Key=%s", header->sec_key);
	if (strlen(header->protocol) > 0) {
		DEBUG("- Sec-WebSocket-Protocol=%s", header->protocol);
		if (strstr(header->protocol, "binary") != NULL) {
			DEBUG("-- accepting binary protocol");
			strcpy(header->protocol, "binary");
		}
	}
	DEBUG("=>SecAccept=%s", header->sec_accept);

	sprintf(header->response,
			"HTTP/1.1 101 Switching Protocols\r\n"
			"Upgrade: websocket\r\n"
			"Connection: Upgrade\r\n"
		    "WebSocket-Origin: %s\r\n"					// origin
		    "WebSocket-Location: ws://%s/tmsremote\r\n"	// host
			"Sec-WebSocket-Accept: %s\r\n"				// sec_accept
			"%s"										// protocol
			"\r\n",
			header->origin,
			header->host,
			header->sec_accept,
			(strcmp(header->protocol, "binary") == 0) ? "Sec-WebSocket-Protocol: binary\r\n" : ""
	);

//    DEBUG("response:\n%s", header->response);
    DEBUG("response:");
    DEBUG("| HTTP/1.1 101 Switching Protocols");
    DEBUG("| Upgrade: websocket");
    DEBUG("| Connection: Upgrade");
    DEBUG("| WebSocket-Origin: %s", header->origin);                 // origin
    DEBUG("| WebSocket-Location: ws://%s/tmsremote", header->host);  // host
    DEBUG("| Sec-WebSocket-Accept: %s", header->sec_accept);         // sec_accept
    if (strcmp(header->protocol, "binary") == 0) {
        DEBUG("| Sec-WebSocket-Protocol: binary");                   // protocol
    }

	return maxlen;

}

/**
 * decodes a websocket frame to payload
 *
 * @param frame		pointer to frame data
 * @param payload	pointer to output data
 * @return payload length
 */
unsigned long ws_decode_frame(unsigned char *frame, unsigned char *payload) {

	if ( (*frame & 0xf0) == 0x80 ) {

		unsigned char *_payload = payload;

		// this is a final frame
		unsigned char opcode = *frame & 0xf;
		frame++;
		unsigned char masked = *frame & 0x80;
		unsigned long len = *frame & 0x7f;
		frame++;
		if (len == 126) {
			// 16 bit payload length
			len = ( (unsigned long) frame[0] << 8 ) | (unsigned long) frame[1];
			frame += 2;
		} else if (len == 127) {
			// 64 bit payload length, we ignore the higher dword
			len =
					//(unsigned long) frame[0] << 56 |
					//(unsigned long) frame[1] << 48 |
					//(unsigned long) frame[2] << 40 |
					//(unsigned long) frame[3] << 32 |
					(unsigned long) frame[4] << 24 |
					(unsigned long) frame[5] << 16 |
					(unsigned long) frame[6] <<  8 |
					(unsigned long) frame[7];
			frame += 8;
		}

		unsigned char mask[] = {0, 0, 0, 0};
		if (masked == 0x80) {
			// get the mask
			mask[0] = frame[0];
			mask[1] = frame[1];
			mask[2] = frame[2];
			mask[3] = frame[3];
			frame += 4;
		}

		// decode the (binary) data
		unsigned long i;
		for ( i = 0; i < len; i++, frame++, _payload++) {
			*_payload = *frame ^ mask[i & 3];
		}

		if ( opcode == wsft_text) {

			// text mode
			*_payload = 0;
			len++;

		} else if (opcode == wsft_close) {

			DEBUG("received closing frame");
			/*DEBUG("mask=%u", masked);
			DEBUG("length=%lu", len);
			for ( i = 0, _payload = payload; i < len; i++, _payload++) {
				DEBUG("%02lu: 0x%02x", i, *_payload);
			}*/

			// overwright received data with close command
			_payload = payload;
			_payload[0] = 8; // Cmd_WsClose;
			_payload[1] = 0;
			_payload[2] = 0;
			_payload[3] = 0;
			len = 4;

		}

		return len;

	}

	return 0;

}

/**
 * creates a websocket frame header (without mask)
 *
 * @param len		length of payload data
 * @param frame		pointer to resulting frame-header data
 * @param opcode	message
 * @param masked	mask the payload? (0: no; 128: yes)
 * @return frame length
 */
unsigned long ws_create_frame_header(unsigned long len, unsigned char *frame, unsigned char opcode, unsigned char masked) {

	unsigned char *ptr = frame;

	*ptr = 0x80 | opcode;
	ptr++;

	// masking
	*ptr = masked;
	if (len < 126) {
		// short frame
		*ptr |= (unsigned char) len;
		ptr++;
	} else if (len < 65536) {
		// medium frame
		ptr[0] |= 126;
		ptr[1] = ( len >> 8 ) & 0xff;
		ptr[2] = ( len      ) & 0xff;
		ptr += 3;
	} else {
		// huge frame
		ptr[0] |= 127;
		ptr[1] = 0;//( len >> 56 ) & 0xff;
		ptr[2] = 0;//( len >> 48 ) & 0xff;
		ptr[3] = 0;//( len >> 40 ) & 0xff;
		ptr[4] = 0;//( len >> 32 ) & 0xff;
		ptr[5] = ( len >> 24 ) & 0xff;
		ptr[6] = ( len >> 16 ) & 0xff;
		ptr[7] = ( len >>  8 ) & 0xff;
		ptr[8] = ( len       ) & 0xff;
		ptr += 9;
	}

	return ptr - frame;

}

/**
 * encodes a payload to a websocket frame
 *
 * @param payload	pointer to payload data
 * @param len		length of data
 * @param frame		pointer to frame data, must be 6/8/14 bytes larger than payload-data (< 126 Byte, < 65536 Bytes, huge)
 * @param opcode	message
 * @param masked	mask the payload? (0: no; 128: yes)
 * @return frame length
 */
unsigned long ws_encode_frame(unsigned char *payload, unsigned long len, unsigned char *frame, unsigned char opcode, int masked) {

	unsigned char *ptr = frame;

	ptr += ws_create_frame_header(len, frame, opcode, masked);

	unsigned long i;

	// mask
	if (masked) {

		unsigned long t = time(NULL);
		unsigned char *mask = ptr;
		ptr[0] = ( t >> 24 ) & 0xff;
		ptr[1] = ( t >> 16 ) & 0xff;
		ptr[2] = ( t >>  8 ) & 0xff;
		ptr[3] = ( t       ) & 0xff;
		ptr += 4;

		// encode the (binary) data with mask
		for ( i = 0; i < len; i++, ptr++, payload++) {
			*ptr = *payload ^ mask[i & 3];
		}

	} else {

		// unmasked
		//memcpy(frame, payload, len);
		//frame += len;
		for ( i = 0; i < len; i++, ptr++, payload++) {
			*ptr = *payload;
		}
	}

	return ptr - frame;

}
