/*
 ============================================================================
 Name        : TMSRemoteTest.c
 Author      : jk
 Version     :
 Copyright   : Your copyright notice
 Description : tests for TMSRemoteMod
 ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include "jk-test.h"
#include "websocket.h"

int test_ws_parse_header() {

	tRequestHeader header;

	char* testheader = "GET /tmsremote HTTP/1.1\r\n"
			"Host: 192.168.1.11:1423\r\n"
			"User-Agent: Mozilla/5.0\r\n"
			"Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8\r\n"
			"Accept-Language: de-de,de;q=0.8,en-us;q=0.5,en;q=0.3\r\n"
			"Accept-Encoding: gzip, deflate\r\n"
			"Sec-WebSocket-Version: 13\r\n"
			"Origin: http://jkserver\r\n"
			"Sec-WebSocket-Key: +BDSoDxuVxtyuCh18zN0qw==\r\n"
			"Sec-WebSocket-Protocol: binary, base64\r\n"
			"Connection: keep-alive, Upgrade\r\n"
			"Pragma: no-cache\r\n"
			"Cache-Control: no-cache\r\n"
			"Upgrade: websocket\r\n\r\n";

	int size = ws_parse_header(testheader , strlen(testheader), &header);

	TEST_ASSERT_GT("result", 0, size);
	TEST_ASSERT_GE("header.ws_version", 6, header.ws_version);
	TEST_ASSERT_STR_EQ("header.sec_accept", "cUB0dMt+a/OlPbmuF6gV663oSdQ=", header.sec_accept);
	TEST_ASSERT_STR_EQ("header.protocol", "binary", header.protocol);

	return TRUE;

}

int test_ws_decode_text_frame() {

	// A single-frame masked text message (contains "Hello")
	unsigned char frame[] = { 0x81, 0x85, 0x37, 0xfa, 0x21, 0x3d, 0x7f, 0x9f, 0x4d, 0x51, 0x58 };

	char msg[1024];

	unsigned long len = ws_decode_frame(frame, (unsigned char*) &msg);

	TEST_ASSERT_EQ("result", 6, len);
	TEST_ASSERT_STR_EQ("msg", "Hello", msg)

	return TRUE;

}

int test_ws_encode_text_frame(int masked) {

	// A single-frame masked text message (contains "Hello")
	char text[] = "Hello";

	unsigned char frame[1024];

	unsigned long len = ws_encode_frame((unsigned char*) &text, 5, frame, wsft_text, masked);

	TEST_ASSERT_EQ("frame length", 7 + ((masked) ? 4 : 0), len);

	char msg[1024];

	ws_decode_frame(frame, (unsigned char*) &msg);
	TEST_ASSERT_STR_EQ("msg", "Hello", msg)

	return TRUE;

}

int main(void) {

	TEST_INIT

	TEST_EXEC( test_ws_parse_header() )

	TEST_EXEC( test_ws_decode_text_frame() )

	TEST_EXEC( test_ws_encode_text_frame(ws_masked) )
	TEST_EXEC( test_ws_encode_text_frame(ws_unmasked) )

	TEST_RESULT

	return EXIT_SUCCESS;

}
